import re
from .. import fragments
from .virus import Virus


HIV1 = Virus(
    virus_name='HIV1',

    strain_name='HIV1',

    supported_commands=[
        'fasta',
        'mutations',
        'patterns',
        'seqreads'
    ],

    default_url='https://hivdb.stanford.edu/graphql',

    gene_defs=[
        {
            'name': 'gag',
            'synonym_pattern': re.compile(r'^\s*(gag|gp0?2)\s*$', re.I)
        },
        {
            'name': 'CA',
            'synonym_pattern': re.compile(
                r'^\s*(ca|capsid)([ _-]?protein)?\s*$', re.I)
        },
        {
            'name': 'pol',
            'synonym_pattern': re.compile(r'^\s*(pol|polyprotein)\s*$', re.I)
        },
        {
            'name': 'PR',
            'synonym_pattern': re.compile(
                r'^\s*(pr|protease)([ _-]?protein)?\s*$', re.I)
        },
        {
            'name': 'RT',
            'synonym_pattern': re.compile(
                r'^\s*(rt|reverse[ -]transcriptase)([ _-]?protein)?\s*$', re.I)
        },
        {
            'name': 'IN',
            'synonym_pattern': re.compile(
                r'^\s*(in|integrase)([ _-]?(pro|protein|proteinase))?\s*$',
                re.I)
        },
        {
            'name': 'vif',
            'synonym_pattern': re.compile(r'^\s*(vif|gp0?3)\s*$', re.I)
        },
        {
            'name': 'vpr',
            'synonym_pattern': re.compile(r'^\s*(vpr|gp0?4)\s*$', re.I)
        },
        {
            'name': 'tat',
            'synonym_pattern': re.compile(r'^\s*(tat|gp0?5)\s*$', re.I)
        },
        {
            'name': 'rev',
            'synonym_pattern': re.compile(r'^\s*(rev|gp0?6)\s*$', re.I)
        },
        {
            'name': 'vpu',
            'synonym_pattern': re.compile(r'^\s*(vpu|gp0?7)\s*$', re.I)
        },
        {
            'name': 'env',
            'synonym_pattern': re.compile(r'^\s*(env|gp0?8)\s*$', re.I)
        },
        {
            'name': 'nef',
            'synonym_pattern': re.compile(r'^\s*(env|gp0?9)\s*$', re.I)
        }
    ],

    default_queries={
        'fasta': fragments.HIV1_SEQUENCE_ANALYSIS_DEFAULT,
        'mutations': fragments.HIV1_MUTATIONS_ANALYSIS_DEFAULT,
        'patterns': fragments.HIV1_MUTATIONS_ANALYSIS_DEFAULT,
        'seqreads': fragments.HIV1_SEQUENCE_READS_ANALYSIS_DEFAULT
    }
)



from __future__ import annotations
import re
from typing import TypedDict, List, Tuple


class CodonReads(TypedDict):
    codon: str
    reads: int


class PosReads(TypedDict):
    gene: str
    position: int
    totalReads: int
    allCodonReads: List[CodonReads]


class UntransRegion(TypedDict):
    name: str
    refStart: int
    refEnd: int
    consensus: str


class SeqReads(TypedDict):
    name: str
    strain: str
    allReads: List[PosReads]
    untranslatedRegions: List[UntransRegion]
    minPrevalence: float
    maxMixtureRate: float
    minCodonReads: int
    minPositionReads: int


class Sequence(TypedDict):
    header: str
    sequence: str


class ServerVer(TypedDict):
    text: str
    publishDate: str


class InputSequence(TypedDict):
    header: str


class Gene(TypedDict):
    name: str


class Mutation(TypedDict):
    consensus: str
    position: int
    AAs: str


class PrettyPairwise(TypedDict):
    positionLine: List[str]
    alignedNAsLine: List[str]


class AlignedGeneSeq(TypedDict):
    gene: Gene
    firstAA: int
    lastAA: int
    prettyPairwise: PrettyPairwise
    alignedNAs: str
    mutations: List[Mutation]


class SequenceResult(TypedDict):
    inputSequence: InputSequence
    alignedGeneSequences: List[AlignedGeneSeq]


class TargetGeneDef(TypedDict):
    name: str
    offset: int
    range: Tuple[int, int]


class _GeneDefRequired(TypedDict):
    # TODO: this hack will be eventually replaced by PEP 655
    name: str
    synonym_pattern: re.Pattern


class GeneDef(_GeneDefRequired, total=False):
    target_genes: List[TargetGeneDef]



import json
import click  # type: ignore
from typing import List, Dict, TextIO, Any

from .. import viruses
from ..sierraclient import SierraClient

from .cli import cli
from .options import url_option, virus_option


@cli.command()
@click.argument('mutations', nargs=-1, required=True)
@url_option('--url')
@virus_option('--virus')
@click.option('-q', '--query', type=click.File('r'),
              help=('A file contains GraphQL fragment definition '
                    'on `MutationsAnalysis`.'))
@click.option('-o', '--output', default='-', type=click.File('w'),
              help='File path to store the JSON result.')
@click.option('--ugly', is_flag=True, help='Output compressed JSON result.')
@click.pass_context
def mutations(
    ctx: click.Context,
    url: str,
    virus: viruses.Virus,
    mutations: List[str],
    query: TextIO,
    output: TextIO,
    ugly: bool
) -> None:
    """
    Run drug resistance and other analysis for PR, RT and/or IN mutations.
    For Example:

    \b
    sierrapy mutations PR:E35E_D RT:T67- IN:M50MI

    Use command "sierrapy patterns" instead if you want to run multiple sets
    of mutations in one request.
    """
    query_text: str
    client: SierraClient = SierraClient(url)
    client.toggle_progress(True)
    if query:
        query_text = query.read()
    else:
        query_text = virus.get_default_query('mutations')
    result: Dict[str, Any] = client.mutations_analysis(mutations, query_text)
    json.dump(result, output, indent=None if ugly else 2)



# -*- coding: utf-8 -*-

import json
from typing import (
    Optional,
    Dict,
    Any,
    Union,
    List,
    Sequence as ListOrTuple,
    Generator,
    Tuple,
    Iterator
)
from more_itertools import chunked

from tqdm import tqdm  # type: ignore
from gql import gql, Client
from gql.transport.requests import RequestsHTTPTransport
from requests.exceptions import HTTPError  # type: ignore
from graphql.language.ast import DocumentNode as gqlDocument

from .common_types import Sequence, SeqReads, ServerVer


VERSION = '0.4.3'
DEFAULT_URL = 'https://hivdb.stanford.edu/graphql'


class ResponseError(Exception):
    pass


class SierraClient:
    url: str
    _client: Optional[Client]
    _progress: bool

    def __init__(self, url: str = DEFAULT_URL):
        self.url = url
        self._client = None
        self._progress = False

    def toggle_progress(self, flag: Union[bool, str] = 'auto') -> None:
        if flag == 'auto':
            self._progress = not self._progress
        else:
            self._progress = bool(flag)

    @property
    def client(self) -> Client:
        if self._client is None:
            transport: RequestsHTTPTransport = \
                RequestsHTTPTransport(self.url, use_json=True, timeout=300)
            transport.headers = {
                'User-Agent': 'sierra-client (python)/{}'.format(VERSION)
            }
            self._client = Client(
                transport=transport,
                fetch_schema_from_transport=True)
        return self._client

    def execute(
        self,
        document: gqlDocument,
        variable_values: Optional[Dict[str, Any]] = None
    ) -> Dict[str, Any]:
        try:
            result: Dict[str, Any] = self.client.execute(
                document, variable_values=variable_values)
            return result
        except HTTPError as e:
            print(e.response.text)
            errors = [
                e['exception']['detailMessage']
                if 'detailMessage' in e['exception']
                else 'Unknown server error'
                for e in e.response.json()['errors']
                if 'exception' in e]
            raise ResponseError(
                'Sierra GraphQL webservice returned errors:\n - ' +
                json.dumps(errors, indent=4))

    def get_introspection(self) -> Dict[str, Any]:
        result: Dict[str, Any] = self.client.introspection
        return result

    def _sequence_analysis(
        self,
        sequences: List[Sequence],
        query: str
    ) -> List[Dict[str, Any]]:
        result: Dict[str, Any] = self.execute(
            gql("""
                query sierrapy($sequences:[UnalignedSequenceInput]!) {{
                    sequenceAnalysis(sequences:$sequences) {{
                        ...F0
                    }}
                }}
                fragment F0 on SequenceAnalysis {{
                    {query}
                }}
                """.format(query=query)),
            variable_values={"sequences": sequences})
        seqResults: List[Dict[str, Any]] = result['sequenceAnalysis']
        return seqResults

    def _pattern_analysis(
        self,
        patterns: ListOrTuple[List[str]],
        pattern_names: ListOrTuple[Optional[str]],
        query: str,
        **kw: Any
    ) -> List[Dict[str, Any]]:
        enable_hivalg: bool = 'algorithms' in kw or 'customAlgorithms' in kw
        extraparams: str = ''
        if enable_hivalg:
            extraparams = ('$algorithms:[ASIAlgorithm] '
                           '$customAlgorithms:[CustomASIAlgorithm]')
        variables: Dict[str, Any] = {
            "patterns": patterns,
            "patternNames": pattern_names
        }
        if enable_hivalg:
            variables['algorithms'] = kw.get('algorithms')
            variables['customAlgorithms'] = kw.get('custom_algorithms')
        result: Dict[str, Any] = self.execute(
            gql("""
                query sierrapy(
                    $patterns:[[String]!]!
                    $patternNames:[String]
                    {extraparams}
                ) {{
                    patternAnalysis(
                        patterns:$patterns
                        patternNames:$patternNames
                    ) {{
                        ...F0
                    }}
                }}
                fragment F0 on MutationsAnalysis {{
                    {query}
                }}
                """.format(query=query, extraparams=extraparams)),
            variable_values=variables)
        patternResults: List[Dict[str, Any]] = result['patternAnalysis']
        return patternResults

    def _sequence_reads_analysis(
        self,
        all_sequence_reads: List[SeqReads],
        query: str
    ) -> List[Dict[str, Any]]:
        result: Dict[str, Any] = self.execute(
            gql("""
                query sierrapy($allSequenceReads:[SequenceReadsInput]!) {{
                    sequenceReadsAnalysis(
                        sequenceReads:$allSequenceReads

                    ) {{
                        ...F0
                    }}
                }}
                fragment F0 on SequenceReadsAnalysis {{
                    {query}
                }}
                """.format(query=query)),
            variable_values={"allSequenceReads": all_sequence_reads})
        seqReadsResults: List[Dict[str, Any]] = result['sequenceReadsAnalysis']
        return seqReadsResults

    def iter_sequence_analysis(
        self,
        sequences: Union[List[Sequence], Iterator[Sequence]],
        query: str,
        step: int = 20
    ) -> Generator[Dict[str, Any], None, None]:
        pbar: Optional[tqdm] = None
        if self._progress:
            pbar = tqdm()
        for partial in chunked(sequences, step):
            yield from self._sequence_analysis(partial, query)
            pbar and pbar.update(len(partial))

    def iter_pattern_analysis(
        self,
        patterns: Iterator[Tuple[str, List[str]]],
        query: str,
        step: int = 20,
        **kw: Any
    ) -> Generator[Dict[str, Any], None, None]:
        pbar: Optional[tqdm] = None
        pats: Tuple[List[str], ...]
        pat_names: Tuple[str, ...]
        if self._progress:
            pbar = tqdm()
        for partial in chunked(patterns, step):
            pat_names, pats = tuple(zip(*partial))
            yield from self._pattern_analysis(
                pats, pat_names, query, **kw
            )
            pbar and pbar.update(step)

    def iter_sequence_reads_analysis(
        self,
        sequence_reads: List[SeqReads],
        query: str,
        step: int = 20
    ) -> Generator[Dict[str, Any], None, None]:
        pbar: Optional[tqdm]
        if self._progress:
            pbar = tqdm(total=len(sequence_reads))
        while sequence_reads:
            yield from self._sequence_reads_analysis(
                sequence_reads[:step], query
            )
            sequence_reads = sequence_reads[step:]
            pbar and pbar.update(step)

    def sequence_analysis(
        self,
        sequences: List[Sequence],
        query: str,
        step: int = 20
    ) -> List[Dict[str, Any]]:
        return list(self.iter_sequence_analysis(sequences, query, step))

    def pattern_analysis(
        self,
        patterns: Iterator[Tuple[str, List[str]]],
        query: str,
        step: int = 20,
        **kw: Any
    ) -> List[Dict[str, Any]]:
        return list(self.iter_pattern_analysis(
            patterns, query, step, **kw
        ))

    def sequence_reads_analysis(
        self,
        sequence_reads: List[SeqReads],
        query: str,
        step: int = 20
    ) -> List[Dict[str, Any]]:
        return list(self.iter_sequence_reads_analysis(
            sequence_reads, query, step
        ))

    def mutations_analysis(
        self,
        mutations: List[str],
        query: str
    ) -> Dict[str, Any]:
        result: Dict[str, Any] = self.execute(
            gql("""
                query sierrapy($mutations:[String]!) {{
                    mutationsAnalysis(mutations:$mutations) {{
                        ...F0
                    }}
                }}
                fragment F0 on MutationsAnalysis {{
                    {query}
                }}
                """.format(query=query)),
            variable_values={"mutations": mutations})
        mutResults: Dict[str, Any] = result['mutationsAnalysis']
        return mutResults

    def current_version(self) -> Tuple[ServerVer, ServerVer]:
        result = self.execute(
            gql("""
                query sierrapy {
                    currentVersion { text, publishDate }
                    currentProgramVersion { text, publishDate }
                }
                """))
        return (result['currentVersion'],
                result.get('currentProgramVersion', {
                    'text': 'Unknown',
                    'publishDate': 'Unknown'
                }))



import json
import click  # type: ignore
from collections import OrderedDict
from typing import (
    List,
    Dict,
    TextIO,
    Optional,
    OrderedDict as tOrderedDict
)

from voluptuous import (  # type: ignore
    Schema, Required, MultipleInvalid, ALLOW_EXTRA
)

from ..common_types import SequenceResult, AlignedGeneSeq

GENES: tOrderedDict = OrderedDict([
    ('PR', 99),
    ('RT', 560),
    ('IN', 288)
])


schema: Schema = Schema([{
    Required('inputSequence'): {
        Required('header'): str
    },
    Required('alignedGeneSequences'): [{
        Required('gene'): {
            Required('name'): str
        },
        Required('firstAA'): int,
        Required('lastAA'): int,
        Required('alignedNAs'): str,
        Required('prettyPairwise'): {
            Required('positionLine'): [str],
            Required('alignedNAsLine'): [str]
        }
    }]
}], extra=ALLOW_EXTRA)


@click.option('--gap-handling', default="hxb2strip",
              type=click.Choice(['squeeze', 'hxb2strip', 'hxb2stripkeepins']),
              help=('Specify how you want the recipe to handle the gaps.\n\n'
                    'Specify "squeeze" to keep every gap in the result '
                    'alignment; "hxb2strip" to strip out non-HXB2 columns; '
                    '"hxb2stripkeepins" to strip not non-HXB2 columns except '
                    'codon insertions.'))
@click.pass_context
def alignment(ctx: click.Context, gap_handling: str) -> None:
    """Export aligned pol sequences from Sierra result."""
    seqheader: str
    concat_seqs: str
    geneseqs: Dict[str, AlignedGeneSeq]
    gene: str
    genesize: int
    geneseq: Optional[AlignedGeneSeq]
    first_aa: int
    last_aa: int
    naseq: List[str]
    posline: List[str]
    naline: List[str]
    pos: str
    nas: str
    naseq_text: str
    output: TextIO = ctx.obj['OUTPUT']
    sequences: List[SequenceResult] = json.load(ctx.obj['INPUT'])
    try:
        schema(sequences)
    except MultipleInvalid as e:
        raise click.ClickException(str(e))
    for seq in sequences:
        seqheader = seq['inputSequence']['header']
        concat_seqs = ''
        geneseqs = {gs['gene']['name']: gs
                    for gs in seq['alignedGeneSequences']}
        for gene, genesize in GENES.items():
            geneseq = geneseqs.get(gene)
            if geneseq:
                first_aa = geneseq['firstAA']
                last_aa = geneseq['lastAA']
                if gap_handling.endswith('keepins'):
                    naseq = []
                    posline = geneseq['prettyPairwise']['positionLine']
                    naline = geneseq['prettyPairwise']['alignedNAsLine']
                    for pos, nas in zip(posline, naline):
                        if not pos.strip() and ' ' in nas:
                            # fs insertions
                            continue
                        naseq.append(nas)
                    naseq_text = ''.join(naseq)
                else:
                    naseq_text = geneseq['alignedNAs']
            else:
                first_aa = 1
                last_aa = genesize
                naseq_text = '.' * genesize * 3
            if gap_handling.startswith('hxb2strip'):
                naseq_text = (
                    ('.' * (first_aa - 1) * 3) +
                    naseq_text +
                    '.' * (genesize - last_aa) * 3
                )
            else:  # gap_handling == 'squeeze'
                raise NotImplementedError()
            concat_seqs += naseq_text
        output.write('>{}\n{}\n'.format(seqheader, concat_seqs))



from typing import List, Set, Dict, Tuple, Optional
from ..common_types import GeneDef, TargetGeneDef


class Virus:
    virus_name: str
    strain_name: str
    supported_commands: List[str]
    default_url: str
    gene_defs: Dict[str, GeneDef]
    source_genes: Set[str]
    ordered_genes: List[str]
    default_queries: Dict[str, str]

    def __init__(
        self,
        virus_name: str,
        strain_name: str,
        supported_commands: List[str],
        default_url: str,
        gene_defs: List[GeneDef],
        default_queries: Dict[str, str]
    ):
        self.virus_name = virus_name
        self.strain_name = strain_name
        self.supported_commands = supported_commands
        self.default_url = default_url
        self.gene_defs = {gdef['name']: gdef for gdef in gene_defs}
        self.source_genes = {
            gdef['name'] for gdef in gene_defs
            if 'target_genes' in gdef
        }
        self.ordered_genes = [gdef['name'] for gdef in gene_defs]
        self.default_queries = default_queries

    def get_default_query(self, command: str) -> str:
        return self.default_queries[command]

    def gene_index(self, gene: str) -> int:
        return self.ordered_genes.index(gene)

    def synonym_to_gene_name(self, gene: str) -> Optional[str]:
        if gene in self.gene_defs:
            return gene

        gene_def: GeneDef
        for gene_def in self.gene_defs.values():
            if gene_def['synonym_pattern'].match(gene):
                return gene_def['name']
        else:
            return None

    def source_gene_to_target_gene_position(
        self, gene: str, pos: int
    ) -> Tuple[Optional[str], Optional[int]]:
        target_gene: TargetGeneDef
        pos_start: int
        pos_end: int
        pos_offset: int
        if gene in self.source_genes:
            for target_gene in self.gene_defs[gene]['target_genes']:
                pos_start, pos_end = target_gene['range']
                pos_offset = target_gene['offset']
                if pos < pos_start:
                    continue
                elif pos > pos_end:
                    continue
                return target_gene['name'], pos + 1 - pos_start + pos_offset
            else:
                return None, None
        else:
            return gene, pos



import os
import re
import json
import math
import tqdm  # type: ignore
import click  # type: ignore
from itertools import chain
from typing import Dict, TextIO, Any, Tuple, Iterator
from more_itertools import chunked

from .. import fastareader, viruses
from ..sierraclient import SierraClient
from ..common_types import Sequence

from .cli import cli
from .options import url_option, virus_option

FASTA_PATTERN = re.compile(r'\.fa(?:s(?:ta)?)?$', re.I)


def iter_fasta_files(
    file_or_dir: Tuple[str, ...]
) -> Iterator[TextIO]:
    for one in file_or_dir:
        if os.path.isfile(one):
            yield open(one)
        else:
            for fn in os.listdir(one):
                if not FASTA_PATTERN.search(fn):
                    continue
                yield open(os.path.join(one, fn))


@cli.command()
@click.argument(
    'fasta',
    nargs=-1,
    type=click.Path(exists=True),
    required=True)
@url_option('--url')
@virus_option('--virus')
@click.option('-q', '--query', type=click.File('r'),
              help=('A file contains GraphQL fragment definition '
                    'on `SequenceAnalysis`.'))
@click.option('-o', '--output', default='-',
              type=click.Path(dir_okay=False),
              help='File path to store the JSON result.')
@click.option('--sharding', type=int, default=100,
              help='Save JSON result files per n sequences.')
@click.option('--no-sharding', is_flag=True,
              help='Save JSON result to a single file.')
@click.option('--step', type=int, default=40,
              help='Send batch requests per n sequences.')
@click.option('--skip', type=int, default=0,
              help='Skip first n sequences.')
@click.option('--total', type=int, default=0,
              help=(
                  'Total number of sequences; '
                  'specify one to visualize a progress bar.'
              ))
@click.option('--ugly', is_flag=True, help='Output compressed JSON result.')
@click.pass_context
def fasta(
    ctx: click.Context,
    url: str,
    virus: viruses.Virus,
    fasta: Tuple[str, ...],
    query: TextIO,
    output: str,
    sharding: int,
    no_sharding: bool,
    step: int,
    skip: int,
    total: int,
    ugly: bool
) -> None:
    """
    Run alignment, drug resistance and other analysis for one or more
    FASTA-format files contained DNA sequences.
    """
    ext: str
    query_text: str
    client: SierraClient = SierraClient(url)
    client.toggle_progress(False)

    fasta_fps: Iterator[TextIO] = iter_fasta_files(fasta)

    sequences: Iterator[Sequence] = chain(*(
        fastareader.load(fp) for fp in fasta_fps
    ))
    idx_offset: int = math.ceil(skip / sharding)
    for _ in zip(range(skip), sequences):
        pass

    if query:
        query_text = query.read()
    else:
        query_text = virus.get_default_query('fasta')

    result: Iterator[
        Dict[str, Any]
    ] = tqdm.tqdm(
        client.iter_sequence_analysis(sequences, query_text, step),
        total=total,
        initial=skip
    )
    if no_sharding:
        with open(output, 'w') as fp:
            json.dump(list(result), fp, indent=None if ugly else 2)
    else:
        output, ext = os.path.splitext(output)
        if not ext:
            ext = 'json'
        for idx, partial in enumerate(chunked(result, sharding)):
            with open('{}.{}{}'.format(
                output, idx + idx_offset, ext
            ), 'w') as fp:
                json.dump(partial, fp, indent=None if ugly else 2)



import re
from .. import fragments
from .virus import Virus


SARS2 = Virus(

    virus_name='SARS2',

    strain_name='SARS2',

    supported_commands=[
        'fasta',
        'mutations',
        'patterns',
        'seqreads'
    ],

    default_url='https://covdb.stanford.edu/sierra-sars2/graphql',

    gene_defs=[
        {
            'name': 'nsp1',
            'synonym_pattern': re.compile(
                r'^\s*nsp1([ _-]?protein)?\s*$', re.I)
        },
        {
            'name': 'nsp2',
            'synonym_pattern': re.compile(
                r'^\s*nsp2([ _-]?protein)?\s*$', re.I)
        },
        {
            'name': 'PLpro',
            'synonym_pattern': re.compile(
                r'^\s*(nsp3|pl|papain-like)'
                r'([ _-]?(pro|protein|proteinase))?\s*$',
                re.I
            )
        },
        {
            'name': 'nsp4',
            'synonym_pattern': re.compile(
                r'^\s*nsp4([ _-]?protein)?\s*$', re.I)
        },
        {
            'name': '_3CLpro',
            'synonym_pattern': re.compile(
                r'^\s*(nsp5|_?3cl|3c-like|mpro|main)'
                r'([ _-]?(pro|protein|proteinase))?\s*$', re.I)
        },
        {
            'name': 'nsp6',
            'synonym_pattern': re.compile(
                r'^\s*nsp6([ _-]?protein)?\s*$', re.I)
        },
        {
            'name': 'nsp7',
            'synonym_pattern': re.compile(
                r'^\s*nsp7([ _-]?protein)?\s*$', re.I)
        },
        {
            'name': 'nsp8',
            'synonym_pattern': re.compile(
                r'^\s*nsp8([ _-]?protein)?\s*$', re.I)
        },
        {
            'name': 'nsp9',
            'synonym_pattern': re.compile(
                r'^\s*nsp9([ _-]?protein)?\s*$', re.I)
        },
        {
            'name': 'nsp10',
            'synonym_pattern': re.compile(
                r'^\s*nsp10([ _-]?protein)?\s*$', re.I)
        },
        {
            'name': 'nsp11',
            'synonym_pattern': re.compile(
                r'^\s*nsp11([ _-]?protein)?\s*$', re.I),
            'target_genes': [
                {
                    'name': 'RdRP',
                    'offset': 0,
                    'range': (1, 9)
                }
            ]
        },
        {
            'name': 'RdRP',
            'synonym_pattern': re.compile(
                r'^\s*(rdrp|rna-dependent rna polymerase)\s*$', re.I)
        },
        {
            'name': 'nsp12',
            'synonym_pattern': re.compile(
                r'^\s*nsp12([ _-]?protein)?\s*$', re.I),
            'target_genes': [
                {
                    'name': 'RdRP',
                    'offset': 9,
                    'range': (1, 923)
                }
            ]
        },
        {
            'name': 'nsp13',
            'synonym_pattern': re.compile(
                r'^\s*nsp13([ _-]?protein)?\s*$', re.I)
        },
        {
            'name': 'nsp14',
            'synonym_pattern': re.compile(
                r'^\s*nsp14([ _-]?protein)?\s*$', re.I)
        },
        {
            'name': 'nsp15',
            'synonym_pattern': re.compile(
                r'^\s*nsp15([ _-]?protein)?\s*$', re.I)
        },
        {
            'name': 'nsp16',
            'synonym_pattern': re.compile(
                r'^\s*nsp16([ _-]?protein)?\s*$', re.I)
        },
        {
            'name': 'ORF1a',
            'synonym_pattern': re.compile(
                r'^\s*orf1a([ _-]?protein)?\s*$', re.I),
            'target_genes': [
                {
                    'name': 'nsp1',
                    'offset': 0,
                    'range': (1, 180)
                },
                {
                    'name': 'nsp2',
                    'offset': 0,
                    'range': (181, 818)
                },
                {
                    'name': 'PLpro',
                    'offset': 0,
                    'range': (819, 2763)
                },
                {
                    'name': 'nsp4',
                    'offset': 0,
                    'range': (2764, 3263)
                },
                {
                    'name': '_3CLpro',
                    'offset': 0,
                    'range': (3264, 3569)
                },
                {
                    'name': 'nsp6',
                    'offset': 0,
                    'range': (3570, 3859)
                },
                {
                    'name': 'nsp7',
                    'offset': 0,
                    'range': (3860, 3942)
                },
                {
                    'name': 'nsp8',
                    'offset': 0,
                    'range': (3943, 4140)
                },
                {
                    'name': 'nsp9',
                    'offset': 0,
                    'range': (4141, 4253)
                },
                {
                    'name': 'nsp10',
                    'offset': 0,
                    'range': (4254, 4392)
                },
                {
                    'name': 'RdRP',
                    'offset': 0,
                    'range': (4393, 4401)
                }
            ]
        },
        {
            'name': 'ORF1b',
            'synonym_pattern': re.compile(
                r'^\s*orf1b([ _-]?protein)?\s*$', re.I),
            'target_genes': [
                {
                    'name': 'RdRP',
                    'offset': 9,
                    'range': (1, 923)
                },
                {
                    'name': 'nsp13',
                    'offset': 9,
                    'range': (924, 1524)
                },
                {
                    'name': 'nsp14',
                    'offset': 9,
                    'range': (1525, 2051)
                },
                {
                    'name': 'nsp15',
                    'offset': 9,
                    'range': (2052, 2397)
                },
                {
                    'name': 'nsp16',
                    'offset': 9,
                    'range': (2398, 2695)
                }
            ]
        },
        {
            'name': 'S',
            'synonym_pattern': re.compile(
                r'^\s*(GU280_gp02|(s|spike|surface)'
                r'([ _-]?(protein|glycoprotein))?)\s*$', re.I)
        },
        {
            'name': 'ORF3a',
            'synonym_pattern': re.compile(
                r'^\s*(GU280_gp03|(orf3a?)([ _-]?protein)?)\s*$', re.I)
        },
        {
            'name': 'E',
            'synonym_pattern': re.compile(
                r'^\s*(GU280_gp04|(orf4|e|envelope|env)([ _-]?protein)?)\s*$',
                re.I
            )
        },
        {
            'name': 'M',
            'synonym_pattern': re.compile(
                r'^\s*(GU280_gp05|(orf5|m|membranes?|mp)([ _-]?protein)?)\s*$',
                re.I
            )
        },
        {
            'name': 'ORF6',
            'synonym_pattern': re.compile(
                r'^\s*(GU280_gp06|orf6([ _-]?protein)?)\s*$',
                re.I
            )
        },
        {
            'name': 'ORF7a',
            'synonym_pattern': re.compile(
                r'^\s*(GU280_gp07|orf7a([ _-]?protein)?)\s*$',
                re.I
            )
        },
        {
            'name': 'ORF7b',
            'synonym_pattern': re.compile(
                r'^\s*(GU280_gp08|orf7b([ _-]?protein)?)\s*$',
                re.I
            )
        },
        {
            'name': 'ORF8',
            'synonym_pattern': re.compile(
                r'^\s*(GU280_gp09|orf8([ _-]?protein)?)\s*$',
                re.I
            )
        },
        {
            'name': 'N',
            'synonym_pattern': re.compile(
                r'^\s*(GU280_gp10|(orf9|n|nucleocapsid|np)'
                r'([ _-]?(protein|phosphoprotein))?)\s*$',
                re.I
            )
        },
        {
            'name': 'ORF10',
            'synonym_pattern': re.compile(
                r'^\s*(GU280_gp11|orf10([ _-]?protein)?)\s*$',
                re.I
            )
        }
    ],

    default_queries={
        'fasta': fragments.SARS2_SEQUENCE_ANALYSIS_DEFAULT,
        'mutations': fragments.SARS2_MUTATIONS_ANALYSIS_DEFAULT,
        'patterns': fragments.SARS2_MUTATIONS_ANALYSIS_DEFAULT,
        'seqreads': fragments.SARS2_SEQUENCE_READS_ANALYSIS_DEFAULT
    }
)



import _csv
import csv
import json
import click  # type: ignore
from collections import OrderedDict
from typing import OrderedDict as tOrderedDict, TextIO, List, Dict, Optional
from voluptuous import (  # type: ignore
    Schema, Required, MultipleInvalid, ALLOW_EXTRA
)

from ..common_types import SequenceResult, AlignedGeneSeq


GENES: tOrderedDict[str, int] = OrderedDict([
    ('PR', 99),
    ('RT', 560),
    ('IN', 288)
])


schema: Schema = Schema([{
    Required('inputSequence'): {
        Required('header'): str
    },
    Required('alignedGeneSequences'): [{
        Required('gene'): {
            Required('name'): str
        },
        Required('mutations'): [{
            Required('consensus'): str,
            Required('position'): int,
            Required('AAs'): str
        }]
    }]
}], extra=ALLOW_EXTRA)


@click.pass_context
def mutationtsv(ctx: click.Context) -> None:
    """Export mutation set of each sequences from Sierra result."""
    seq: SequenceResult
    seqheader: str
    geneseqs: Dict[str, AlignedGeneSeq]
    row: List[str]
    gene: str
    geneseq: Optional[AlignedGeneSeq]
    mutations: str
    output: TextIO = ctx.obj['OUTPUT']
    sequences: List[SequenceResult] = json.load(ctx.obj['INPUT'])
    try:
        schema(sequences)
    except MultipleInvalid as e:
        raise click.ClickException(str(e))
    writer: _csv._writer = csv.writer(output, delimiter='\t')
    writer.writerow(['Header'] + ['{} Mutations'.format(gene)
                                  for gene in GENES.keys()])
    for seq in sequences:
        seqheader = seq['inputSequence']['header']
        geneseqs = {gs['gene']['name']: gs
                    for gs in seq['alignedGeneSequences']}
        row = [seqheader]
        for gene, _ in GENES.items():
            geneseq = geneseqs.get(gene)
            mutations = ''
            if geneseq:
                mutations = ', '.join([
                    '{consensus}{position}{AAs}'
                    .format(**m).replace('-', 'Deletion')
                    for m in geneseq['mutations']])
            row.append(mutations)
        writer.writerow(row)



#! /usr/bin/env python
# -*- coding: UTF-8 -*-

import os
import setuptools  # type: ignore

from typing import List

VERSION = '0.4.3'


def strip_comments(line: str) -> str:
    if line.startswith('-i '):
        return ''
    return line.split('#', 1)[0].strip()


def req(filename: str) -> List[str]:
    ln: str
    requires: set
    with open(os.path.join(os.getcwd(), filename)) as fp:
        requires = set([strip_comments(ln) for ln in fp.readlines()])
        requires -= set([''])
    return list(requires)


setup_params = dict(
    name="sierrapy",
    version=VERSION,
    url="https://github.com/hivdb/sierra-client/tree/master/python",
    author='Philip Tzou',
    author_email="philiptz@stanford.edu",
    description='A Client of HIVdb Sierra GraphQL Webservice.',
    packages=['sierrapy',
              'sierrapy/fragments',
              'sierrapy/recipes',
              'sierrapy/commands',
              'sierrapy/viruses'],
    install_requires=req('requirements.txt'),
    # tests_require=reqs('test-requirements.txt'),
    include_package_data=True,
    entry_points={'console_scripts': [
        'sierrapy = sierrapy.cmds:main',
    ]},
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Topic :: Scientific/Engineering :: Bio-Informatics'],
    # test_suite="nose.collector",
    zip_safe=True)

if __name__ == '__main__':
    setuptools.setup(**setup_params)



import os
import re
import click  # type: ignore
from typing import Any, Callable, List

from .. import viruses


def url_option_callback(
    ctx: click.Context,
    param: click.Option,
    value: str
) -> str:
    if 'URL' in ctx.obj:
        value = ctx.obj['URL']
    elif value == '__default_url__':
        value = ctx.params['virus'].default_url
    else:
        ctx.obj['URL'] = value
    return value


def url_option(*args: Any) -> Callable:
    func: Callable = click.option(
        *args,
        default='__default_url__',
        callback=url_option_callback,
        help=(
            'URL of Sierra GraphQL Web Service.  '
            '[default: production URL varied by virus]'
        ))
    return func


def virus_option_callback(
    ctx: click.Context,
    param: click.Option,
    value: str
) -> viruses.Virus:
    virus: viruses.Virus
    if 'virus' in ctx.obj:
        virus = ctx.obj['virus']
        return virus
    virus = getattr(viruses, value or 'HIV1')
    if 'fasta' in virus.supported_commands:
        from . import fasta  # noqa
    elif ctx.command.name == 'fasta':
        raise click.UsageError(
            f"Command 'fasta' is not supported by --virus={value}."
        )
    if 'mutations' in virus.supported_commands:
        from . import mutations  # noqa
    elif ctx.command.name == 'mutations':
        raise click.UsageError(
            f"Command 'mutations' is not supported by --virus={value}."
        )
    if 'patterns' in virus.supported_commands:
        from . import patterns  # noqa
    elif ctx.command.name == 'patterns':
        raise click.UsageError(
            f"Command 'patterns' is not supported by --virus={value}."
        )
    if 'seqreads' in virus.supported_commands:
        from . import seqreads  # noqa
    elif ctx.command.name == 'seqreads':
        raise click.UsageError(
            f"Command 'seqreads' is not supported by --virus={value}."
        )
    if value is not None:
        ctx.obj['virus'] = virus
    return virus


def virus_option(*args: Any) -> Callable:
    func: Callable = click.option(
        *args,
        type=click.Choice(['HIV1', 'HIV2', 'SARS2']),
        default='HIV1', show_default=True,
        is_eager=True,
        callback=virus_option_callback,
        help='Specify virus to be analyzed.')
    return func


def file_or_dir_argument(*args: Any, pattern: re.Pattern) -> Callable:

    def file_or_dir_callback(
        ctx: click.Context,
        param: click.Argument,
        value: str
    ) -> List[str]:
        if pattern.search(value):
            return [value]
        else:
            new_value: List[str] = []
            for dirpath, _, filenames in os.walk(value, followlinks=True):
                for filename in filenames:
                    if pattern.search(filename):
                        new_value.append(os.path.join(dirpath, filename))
            return new_value

    func: Callable = click.argument(
        *args,
        type=click.Path(exists=True, file_okay=True,
                        dir_okay=True, resolve_path=True),
        callback=file_or_dir_callback,
        required=True)
    return func



# -*- coding: utf-8 -*-
import os
import re
import json
import math
import tqdm  # type: ignore
import click  # type: ignore
from more_itertools import chunked
from typing import List, Dict, Any, TextIO, Optional, Iterator, Tuple

from .. import viruses
from ..sierraclient import SierraClient

from .cli import cli
from .options import url_option, virus_option


def iter_patterns(
    pattern_files: List[TextIO]
) -> Iterator[Tuple[str, List[str]]]:
    fp: TextIO
    ptn_name: Optional[str]
    for fp in pattern_files:
        ptn_name = None
        for line in fp:
            ptn = line.strip()
            if not ptn:
                continue
            if line.startswith('#'):
                continue
            if line.startswith('>'):
                ptn_name = line[1:].strip()
                continue
            elif ptn_name is None:
                ptn_name = ptn
            yield ptn_name, re.split(r'[,;+ \t]+', ptn)
            ptn_name = None


@cli.command()
@click.argument('patterns', nargs=-1, required=True, type=click.File('r'))
@url_option('--url')
@virus_option('--virus')
@click.option('-q', '--query', type=click.File('r'),
              help=('A file contains GraphQL fragment definition '
                    'on `MutationsAnalysis`.'))
@click.option('-o', '--output', default='-',
              type=click.Path(dir_okay=False),
              help='File path to store the JSON result.')
@click.option('--sharding', type=int, default=100,
              help='Save JSON result files per n patterns.')
@click.option('--no-sharding', is_flag=True,
              help='Save JSON result to a single file.')
@click.option('--step', type=int, default=40,
              help='Send batch requests per n patterns.')
@click.option('--skip', type=int, default=0,
              help='Skip first n patterns.')
@click.option('--total', type=int, default=0,
              help=(
                  'Total number of patterns; '
                  'specify one to visualize a progress bar.'
              ))
@click.option('--ugly', is_flag=True, help='Output compressed JSON result.')
@click.pass_context
def patterns(
    ctx: click.Context,
    url: str,
    virus: viruses.Virus,
    patterns: List[TextIO],
    query: TextIO,
    output: str,
    sharding: int,
    no_sharding: bool,
    step: int,
    skip: int,
    total: int,
    ugly: bool
) -> None:
    """
    Run drug resistance and other analysis for one or more files contains
    lines of PR, RT and/or IN mutations based on HIV-1 type B consensus.
    Each line is treated as a unique pattern. For example:

    \b
    >set1
    RT:M41L + RT:M184V + RT:L210W + RT:T215Y
    >set2
    PR:L24I + PR:M46L + PR:I54V + PR:V82A

    The following delimiters are supported: commas (,), plus signs (+),
    semicolon(;), whitespaces and tabs. The consensus sequences can be
    retrieved from HIVDB website: <https://goo.gl/ZBthkt>.
    """
    client: SierraClient = SierraClient(url)
    client.toggle_progress(False)

    fp: TextIO
    query_text: str
    ptns: Iterator[Tuple[str, List[str]]] = iter_patterns(patterns)
    idx_offset: int = math.ceil(skip / sharding)
    for _ in zip(range(skip), ptns):
        pass

    if query:
        query_text = query.read()
    else:
        query_text = virus.get_default_query('patterns')

    result: Iterator[
        Dict[str, Any]
    ] = tqdm.tqdm(
        client.iter_pattern_analysis(ptns, query_text, step),
        total=total,
        initial=skip
    )
    if no_sharding:
        with open(output, 'w') as fp:
            json.dump(list(result), fp, indent=None if ugly else 2)
    else:
        output, ext = os.path.splitext(output)
        if not ext:
            ext = 'json'
        for idx, partial in enumerate(chunked(result, sharding)):
            with open('{}.{}{}'.format(
                output, idx + idx_offset, ext
            ), 'w') as fp:
                json.dump(partial, fp, indent=None if ugly else 2)
